<?php

namespace NoccyLabs\JsonEdit\Editor;

use Exception;
use NoccyLabs\JsonEdit\List\TreeList;
use NoccyLabs\JsonEdit\Settings;
use NoccyLabs\JsonEdit\Terminal\Terminal;
use NoccyLabs\JsonEdit\Tree\ArrayNode;
use NoccyLabs\JsonEdit\Tree\CollapsibleNode;
use NoccyLabs\JsonEdit\Tree\ObjectNode;
use NoccyLabs\JsonEdit\Tree\Tree;
use NoccyLabs\JsonEdit\Tree\ValueNode;
use Symfony\Component\Yaml\Exception\ParseException;
use Symfony\Component\Yaml\Yaml;

class Editor
{

    public static string $helpText;

    private Tree $document;

    private TreeList $list;

    private int $currentRow = 0;

    private int $scrollOffset = 0;

    private ?string $filename = "untitled.json";

    private ?string $shortfilename = "untitled.json";

    private bool $running = true;

    private bool $modified = false;

    /**
     * Constructor
     *
     * @param Terminal $term
     */
    public function __construct(private Terminal $term)
    {
        $this->document = new Tree();

        $this->list = new TreeList($this->document);

        $this->setWindowTitle($this->shortfilename." - JSONEdit");
    }

    /**
     * Read a file into the Tree and TreeList
     *
     * @param string $filename
     * @return void
     */
    public function loadFile(string $filename): void
    {
        $ext = strtolower(pathinfo($filename, PATHINFO_EXTENSION));
        switch ($ext) {
            case 'json':
                $doc = json_decode(file_get_contents($filename));
                break;
            case 'yml':
            case 'yaml':
                $doc = Yaml::parseFile($filename);
                break;    
            default:
                throw new \RuntimeException("Unable to read file of type {$ext}");
        }
        $this->setOpenedFilename($filename);
        $this->document->load($doc);
        $this->list->parseTree();
    }

    /**
     * Load a document from array/object
     *
     * @param mixed $document
     * @return void
     */
    public function loadDocument(mixed $document): void
    {
        $this->document->load($document);
        $this->list->parseTree();
    }

    /**
     * Run the editor
     *
     * @return void
     */
    public function run()
    {
        $this->list->parseTree();
        $this->redrawEditor();

        while ($this->running) {
            [$w,$h] = $this->term->getSize();
            if ($this->term->resized) {
                $this->redrawEditor();
                $this->term->resized = false;
            }
            usleep(10000);
            $read = $this->term->readKey();
            switch ($read) {
                case 'k{UP}':
                    $this->currentRow = max(0, $this->currentRow - 1);
                    $this->redrawEditor();
                    break;
                case 'k{DOWN}':
                    $this->currentRow = min(count($this->list), $this->currentRow + 1);
                    $this->redrawEditor();
                    break;
                case 'k{PGUP}':
                    $this->currentRow = max(0, $this->currentRow - ($h - 3));
                    $this->redrawEditor();
                    break;
                case 'k{PGDN}':
                    $this->currentRow = min(count($this->list), $this->currentRow + ($h - 3));
                    $this->redrawEditor();
                    break;
                case 'k{HOME}':
                    $this->currentRow = 0;
                    $this->redrawEditor();
                    break;
                case 'k{END}':
                    $this->currentRow = count($this->list);
                    $this->redrawEditor();
                    break;

                case 'E':
                    $this->doEditKey();
                    break;
    
                case 'e':
                    $this->doEditValue();
                    break;

                case '[':
                    break;                    

                case 'D':
                    $this->doDeleteValue();
                    break;

                // FIXME make sure this clones; editing a clone updates original as well
                case "\x04": // ctrl-d
                    $node = $this->list->getNodeForIndex($this->currentRow);
                    $coll = $this->list->findNearestCollection($this->currentRow, true);
                    if (!$coll)
                        break;
                    $collNode = $this->list->getNodeForIndex($coll);
                    if ($collNode instanceof ArrayNode) {
                        $collNode->append(clone $node);
                        $this->modified = true;
                        $this->list->parseTree();
                        $this->redrawEditor();
                    } else {
                        $this->showMessage("\e[97;41mCan only duplicate in arrays");
                    }
                    break;

                case 'I':
                    $this->doInsertMenu();
                    break;

                case 'H':
                    $this->doShowAbout();
                    break;

                case 'h':
                    $this->doShowHelp();
                    break;

                case 'i':
                    $this->doInsertValue();
                    $this->modified = true;
                    break;

                case "\x18": // ctrl-x
                    $this->doExit();
                    break;

                case "g":
                    Settings::$indentationGuides = !Settings::$indentationGuides;
                    $this->redrawEditor();
                    $this->showMessage("Toggle indentation guides (g)");
                    break;

                case "b":
                    Settings::$collapseBefore = !Settings::$collapseBefore;
                    $this->redrawEditor();
                    $this->showMessage("Toggle collapse icon before item (b)");
                    break;

                case "t":
                    Settings::$tailLine = !Settings::$tailLine;
                    $this->redrawEditor();
                    $this->showMessage("Toggle tail line (t)");
                    break;
    
                case "+":
                    $node = $this->list->getNodeForIndex($this->currentRow);
                    if ($node instanceof CollapsibleNode) {
                        $node->collapse(!$node->isCollapsed());
                        $this->list->parseTree();
                        $this->redrawEditor();
                    }
                    break;

                case "-":
                    $this->doCollapseAll();
                    break;
                        
                case "k{LEFT}":
                    $node = $this->list->getNodeForIndex($this->currentRow);
                    if ($node instanceof CollapsibleNode) {
                        if ($node->isCollapsed()) {
                            $path = $this->list->getPathForIndex($this->currentRow);
                            $parent = $this->list->getIndexForPath(dirname($path));
                            if ($parent) {
                                $this->currentRow = $parent;
                            }
                        }
                        $node->collapse(true);
                        $this->list->parseTree();
                        $this->redrawEditor();
                    } else {
                        $path = $this->list->getPathForIndex($this->currentRow);
                        $parent = $this->list->getIndexForPath(dirname($path));
                        if ($parent) {
                            $this->currentRow = $parent;
                        }
                        $this->redrawEditor();
                    }
                    break;
    
                case "k{RIGHT}":
                    $node = $this->list->getNodeForIndex($this->currentRow);
                    if ($node instanceof CollapsibleNode) {
                        $node->collapse(false);
                        $this->list->parseTree();
                        $this->redrawEditor();
                    }
                    break;
    
                case "q":
                    Settings::$editorQuotedKeys = !Settings::$editorQuotedKeys;
                    $this->redrawEditor();
                    $this->showMessage("Toggle quoted keys (q)");
                    break;

                case "c":
                    Settings::$compactGroups = !Settings::$compactGroups;
                    $this->list->parseTree();
                    $this->redrawEditor();
                    $this->showMessage("Toggle compact groups (c)");
                    break;

                case "/":
                    $this->doSearch();
                    break;

                case "\x0d": // enter
                    $node = $this->list->getNodeForIndex($this->currentRow);
                    if ($node instanceof ValueNode) {
                        if ($node->value === true) {
                            $node->value = false;
                            $this->modified = true;
                            $this->redrawEditor();
                        } elseif ($node->value === false) {
                            $node->value = true;
                            $this->modified = true;
                            $this->redrawEditor();
                        } else {
                            $this->doEditValue();
                        }
                    }
                    break;

                case "\x03": // ctrl-c
                    $this->showMessage("\e[30;43mPress Ctrl-X to exit.");
                    break;
    
                case "\x0e": // ctrl-n
                    if ($this->modified) {
                        $menu = new Menu($this->term, [
                            'cancel' => "Return to editor",
                            'save' => "Save changes",
                            'discard' => "Discard changes",
                            //'sep0' => "---",
                            //'__paste' => "Paste from clipboard",
                        ], 'Document is modified');
                        $this->redrawInfoBar([ '^C' => 'Cancel', '↑/↓' => 'Select option', 'Enter' => 'Accept' ]);
                        $sel = $menu->display(0, 0, 30, 0, "value");
                        $this->redrawEditor();
                        switch ($sel) {
                            case 'cancel':
                                break(2);
                            case 'save':
                                if (!$this->doWriteFile()) {
                                    break(2);
                                }
                                break;
                            case 'discard':
                                break;
                        }
                        $this->modified = false;
                        
                    }

                    $this->document->load((object)[]);
                    $this->list->parseTree();
                    $this->filename = "untitled.json";
                    $this->shortfilename = "untitled.json";
                    $this->modified = false;
                    $this->redrawEditor();
                    break;

                case "\x13": // ctrl-s
                    $this->doSettingsMenu();
                    break;
                        
                case "\x0F": // ctrl-o
                    $this->doOpenFile();
                    break;

                case "\x12": // ctrl-r
                    $this->doReadFile();
                    break;

                case "\x17": // ctrl-w
                    $this->doWriteFile();
                    break;

                case null:
                    break;
                
                default:
                    $this->term->setCursor(1, $h, true);
                    echo sprintf("%s %02x %03d", ctype_print($read)?$read:'.', ord($read), ord($read));
                    sleep(1);
            }
        }

        Settings::save(SETTINGS_FILE);

    }

    public function doSearch(): void
    {

        $search = $this->ask("\e[36mSearch: ");

        if ($search === null) {
            $this->redrawEditor();
            return;
        }

        if ($search === '') {
            $this->list->searchResults = [];
        } else {
            $this->list->searchResults = $this->document->search($search);
        }

        $this->redrawEditor();
    }

    /**
     * Handler for read file command (ctrl-R)
     *
     * @return void
     */
    private function doReadFile(): void
    {
        [$w,$h] = $this->term->getSize();

        $readFrom = $this->ask("\e[33mRead from:\e[0m ", "");
        if ($readFrom === null) {
            $this->redrawInfoBar();
            return;
        }
        if (!file_exists($readFrom)) {
            $this->showMessage("\e[97;41mFile does not exist");
            return;
        }
        if (!is_readable($readFrom)) {
            $this->showMessage("\e[97;41mFile not readable");
            return;
        }
        $ext = strtolower(pathinfo($readFrom, PATHINFO_EXTENSION));
        switch ($ext) {
            case 'json':
                $doc = json_decode(file_get_contents($readFrom));
                break;
            case 'yml':
            case 'yaml':
                $doc = Yaml::parseFile($readFrom);
                break;
            default:
                $this->showMessage("\e[97;41mUnable to read format: {$ext}");
                return;
        }

        $this->filename = $readFrom;
        $this->shortfilename = basename($readFrom);

        $this->document->load($doc);
        $this->currentRow = 0;
        $this->list->parseTree();
        $this->redrawEditor();

        $this->setWindowTitle($this->shortfilename." - JSONEdit");

        $this->showMessage("\e[97;42mLoaded {$readFrom}");
    }

    /**
     * Handler for the write file command (ctrl-W)
     *
     * @return bool
     */
    private function doWriteFile(): bool
    {
        [$w,$h] = $this->term->getSize();

        $saveTo = $this->ask("\e[33mWrite to:\e[0m ", $this->filename);
        if (!$saveTo) {
            return false;
        }
        if (file_exists($saveTo)) {
            $overwrite = (new Menu($this->term, [
                'overwrite' => "Overwrite the file",
                'abort' => "Abort"
            ], "The file exists"))->display(0, 0, 30, 0, '');
            if ($overwrite == 'abort') {
                $this->redrawEditor();
                return false;
            }
        }

        $doc = $this->document->save();


        $ext = strtolower(pathinfo($saveTo, PATHINFO_EXTENSION));
        switch ($ext) {
            case 'json':
                file_put_contents($saveTo, json_encode($doc, JSON_PRETTY_PRINT|JSON_UNESCAPED_SLASHES)."\n");
                break;
            case 'yml':
            case 'yaml':
                $doc = json_decode(json_encode($doc), true);
                file_put_contents($saveTo, Yaml::dump($doc));
                break;
            default:
                $this->showMessage("\e[97;41mUnable to write format: {$ext}");
                return false;
        }

        $this->setOpenedFilename($saveTo);
        $this->modified = false;
        $this->redrawEditor();

        $this->showMessage("\e[97;42mWrote to {$saveTo}");
        return true;

    }

    private function doOpenFile()
    {
        $includeHidden = false;
        $wd = dirname($this->filename);
        if ($wd == '.') $wd = getcwd();
        $menu = new Menu($this->term, []);
        while (true) {
            $items = [
                dirname($wd) => sprintf(
                    "%-30s %10s %20s",
                    "<..>",
                    "<DIR>",
                    "-"
                )
            ];
            if ($includeHidden) {
                $files = array_merge(glob($wd.'/.[!.]*'), glob($wd.'/*'));
            }else {
                $files = glob($wd."/*");
            }
            // do two sweeps, to sort directories.
            foreach ($files as $file) {
                if (!is_dir($file)) continue;
                $items[$file] = sprintf(
                    "%-30s %10s %20s",
                    mb_substr("<".basename($file).">", 0, 30),
                    "<DIR>",
                    date("Y-m-d H:i:s", filemtime($file))
                );
            }
            foreach ($files as $file) {
                if (is_dir($file)) continue;
                $items[$file] = sprintf(
                    "%-30s %10s %20s",
                    mb_substr(basename($file), 0, 30),
                    filesize($file),
                    date("Y-m-d H:i:s", filemtime($file))
                );
            }
            $this->redrawEditor();
            $this->redrawInfoBar([ '^C' => "Cancel", '^H' => "Hidden", '↑↓' => "Navigate", 'Enter' => "Select/Open" ]);
            $sel = $menu
                ->setKeyhandler(function ($key) use (&$includeHidden) {
                    if ($key == "\x08") {
                        $includeHidden = !$includeHidden;
                        return true;
                    }
                })
                ->setItems($items)
                ->setTitle($wd)
                ->display(0, 0, 70, 0, $this->filename);
            if ($sel === null) {
                $this->redrawEditor();
                return;
            }
            if (is_dir($sel)) {
                $wd = $sel;
            }
            if (is_file($sel)) {
                $body = file_get_contents($sel);
                $json = json_decode($body);
                if ($json !== null) {
                    $this->loadDocument($json);
                    break;
                }
                try {
                    $yaml = Yaml::parse($body);
                    $this->loadDocument($yaml);
                    break;
                } catch (ParseException $e) {
                    $this->showMessage("\e[41;93mUnsupported file format");
                }
            }
        }
        $this->setOpenedFilename($sel);
        $this->redrawEditor();
        $this->showMessage("\e[42;97mOpened {$this->shortfilename}");
    }

    private function doShowHelp(): void
    {
        [$w,$h] = $this->term->getSize();

        $width = min(90, $w - 10);
        $height = min(40, $h - 6);
        $left = (int)round(($w / 2) - ($width / 2));
        $top = (int)round(($h / 2) - ($height / 2));

        $msg = new MessageBox($this->term, self::$helpText, "Help");
        $this->redrawInfoBar([ '↑/↓' => 'Scroll', '^C' => 'Close' ]);
        $msg->display($left, $top, $width, $height);
        $this->redrawEditor();

    }

    private function doShowAbout(): void
    {
        [$w,$h] = $this->term->getSize();

        $text = <<<EOT
        ## JSONEdit Alpha - Copyright (C) 2024, NoccyLabs

        Licensed under GNU GPL v3.0 or later.

        This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

        This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

        EOT."\n/ Version: ".APP_VERSION."\n/ Built at: ".APP_BUILDDATE;

        $width = 60;
        $height = 0; // 22;
        $left = 0; // (int)round(($w / 2) - ($width / 2));
        $top = 0; // (int)round(($h / 2) - ($height / 2));

        $msg = new MessageBox($this->term, $text, "About JSONEdit ".APP_VERSION);
        $this->redrawInfoBar([ '↑/↓' => 'Scroll', '^C' => 'Close' ]);
        $msg->display($left, $top, $width, $height);
        $this->redrawEditor();

    }

    /**
     * Edit selected key
     *
     * @return void
     */
    private function doEditKey(): void
    {
        [$w,$h] = $this->term->getSize();

        //$coll = $this->list->findNearestCollection($this->currentRow);
        $entry = $this->list->getEntryForIndex($this->currentRow);
        $path = $this->list->getPathForIndex($this->currentRow);
        $parentIndex = $this->list->getIndexForPath(dirname($path));
        $parent = $this->list->getEntryForIndex($parentIndex);
        
        if ($parent->node instanceof ObjectNode) {
            $newVal = $this->ask("\e[0;33mnew key:\e[0m ", $entry->key);
            if (!empty($newVal)) {
                try {
                    $parent->node->rename($entry->key, $newVal);
                    $entry->key = $newVal;
                    $this->redrawEditor();
                } catch (Exception $e) {
                    $this->redrawEditor();
                    $this->showMessage("\e[97;41m".$e->getMessage());
                }
            }
        } else {
            // $this->term->setCursor(1, $h);
            // echo "\e[97;41mCan only edit keys on objects\e[K\e[0m";
            $this->showMessage("\e[97;41mCan only edit keys on objects");
        }
        
    }

    /**
     * Edit selected value
     *
     * @return void
     */
    private function doEditValue(): void
    {
        [$w,$h] = $this->term->getSize();

        //$coll = $this->list->findNearestCollection($this->currentRow);
        $node = $this->list->getNodeForIndex($this->currentRow);
        if ($node instanceof ValueNode) {
            $val = json_encode($node->value, JSON_UNESCAPED_SLASHES);
            $newVal = $this->ask("\e[33;1mnew value:\e[0m ", $val);
            // We get null on ctrl-C, so non-null is input
            if ($newVal !== null) {
                // Break out if the value has not changed
                if ($newVal === $val) {
                    $this->redrawInfoBar();
                    return;
                }
                // Attempt to decode the value
                $val = json_decode($newVal);
                // If the string decodes to null, but isn't 'null', treat it as a string
                if ($val === null && $newVal !== 'null') {
                    $val = $newVal;
                }
                $node->value = $val;
                $this->modified = true;
                $this->redrawEditor();
            } else {
                $this->redrawInfoBar();
            }
        } else {
            $this->showMessage("\e[97;41mCan not edit array/object");
        }
    }

    private function doDeleteValue(): void
    {
        $node = $this->list->getNodeForIndex($this->currentRow);
        $path = $this->list->getPathForIndex($this->currentRow);
        $parentPath = dirname($path);
        $deleteKey = basename($path);
        $collNode = $this->list->getNodeForPath($parentPath);
        if ($collNode instanceof ArrayNode) {
            $collNode->removeIndex(intval($deleteKey));
            $this->list->parseTree();
            $this->redrawEditor();
            $this->modified = true;
        } elseif ($collNode instanceof ObjectNode) {
            $collNode->unset($deleteKey);
            $this->list->parseTree();
            $this->redrawEditor();
            $this->modified = true;
        } else {
            $this->showMessage("\e[97;41mCan only delete from object, array");
        }
    }

    /**
     * Insert a new value
     *
     * @param mixed $value
     * @return void
     */
    private function doInsertValue(mixed $value = null): void
    {
        $coll = $this->list->findNearestCollection($this->currentRow);
        $node = $this->list->getNodeForIndex($coll);
        if ($node instanceof ObjectNode) {
            $key = $this->ask("\e[0;33mkey:\e[0m ");
            if (empty($key)) {
                $this->redrawInfoBar();
                return;
            }
            if ($node->hasKey($key)) {
                $this->showMessage("\e[97;41mThe key {$key} already exists");
                return;
            }
        }

        if ($value === null)
            $value = $this->ask("\e[0;32mvalue:\e[0m ");
        if ($value !== null) {
            $newvalue = json_decode($value);
            // If the string decodes to null, but isn't 'null', treat it as a string
            if ($newvalue === null && $value !== 'null') {
                $newvalue = $value;
            }
            $value = $newvalue;

            $valueNode = match (true) {
                is_array($value) => new ArrayNode([]),
                is_object($value) => new ObjectNode([]),
                default => new ValueNode($value)
            };
            if ($node instanceof ArrayNode) {
                $node->append($valueNode);
            } elseif ($node instanceof ObjectNode) {
                $node->set($key, $valueNode);
            }
            $this->modified = true;
            $this->list->parseTree();
            $this->redrawEditor();
        } else {
            $this->redrawInfoBar();
        }
    }

    public function doInsertMenu(): void
    {
        $menu = new Menu($this->term, [
            'value' => "Insert Value",
            'object' => "Insert Object{}",
            'array' => "Insert Array[]",
        ], 'Insert');
        $this->redrawInfoBar([ '^C' => 'Cancel', '↑/↓' => 'Select option', 'Enter' => 'Accept' ]);
        $sel = $menu->display(0, 0, 40, 0, "value");
        $this->redrawEditor();
        switch ($sel) {
            case 'value':
                $this->doInsertValue();
                $this->modified = true;
                break;
            case 'array':
                $this->doInsertValue('[]');
                $this->modified = true;
                break;
            case 'object':
                $this->doInsertValue('{}');
                $this->modified = true;
                break;
        }
    }

    public function doCollapseAll(): void
    {
        foreach ($this->list as $path => $entry) {
            $node = $entry->node;
            if ($path == "/") {
                if ($node instanceof CollapsibleNode) {
                    $node->collapse(false);
                }
            } else {
                if ($node instanceof CollapsibleNode) {
                    $node->collapse(true);
                }
            }
        }
        $this->currentRow = 0;
        $this->list->parseTree();
        $this->redrawEditor();
    }

    public function doExit(): void
    {
        if ($this->modified) {
            $menu = new Menu($this->term, [
                'cancel' => "Return to editor",
                'save' => "Save changes",
                'discard' => "Discard changes",
            ], 'Document is modified');
            $this->redrawInfoBar([ '^C' => 'Cancel', '↑/↓' => 'Select option', 'Enter' => 'Accept' ]);
            $sel = $menu->display(0, 0, 30, 0, "value");
            $this->redrawEditor();
            switch ($sel) {
                case 'cancel':
                    break;
                case 'save':
                    if ($this->doWriteFile()) {
                        $this->running = false;
                    }
                    break;
                case 'discard':
                    $this->running = false;
                    break;
            }
        } else {
            $this->running = false;
        }
    }

    public function doSettingsMenu(): void
    {
        $sel = null;
        $mark = "\u{f657} ";
        $clear = "\u{f630} ";

        do {
            $this->redrawEditor();
            $this->redrawInfoBar([ "^C" => "Close", "↑↓" => "Navigate", "Enter" => "Toggle/Select" ]);
            $items = [
                'collapseBefore' => (Settings::$collapseBefore?$mark:$clear)."Collapse before key",
                'compactGroups' => (Settings::$compactGroups?$mark:$clear)."Compact groups",
                'quotedKeys' => (Settings::$editorQuotedKeys?$mark:$clear)."Quoted keys",
                'tailLine' => (Settings::$tailLine?$mark:$clear)."Show tail line",
                'indentationGuides' => (Settings::$indentationGuides?$mark:$clear)."Show indentation guides",
                'highlightIndices' => (Settings::$highlightIndices?$mark:$clear)."Highlight array indices",
            ];
            $sel = (new Menu($this->term, $items, "Settings"))
                ->display(0, 0, 40, 0, $sel);
            switch ($sel) {
                case 'collapseBefore':
                    Settings::$collapseBefore = !Settings::$collapseBefore;
                    break;
                case 'compactGroups':
                    Settings::$compactGroups = !Settings::$compactGroups;
                    $this->list->parseTree();
                    break;
                case 'quotedKeys':
                    Settings::$editorQuotedKeys = !Settings::$editorQuotedKeys;
                    break;
                case 'tailLine':
                    Settings::$tailLine = !Settings::$tailLine;
                    break;
                case 'indentationGuides':
                    Settings::$indentationGuides = !Settings::$indentationGuides;
                    break;
                case 'highlightIndices':
                    Settings::$highlightIndices = !Settings::$highlightIndices;
                    break;
            }
        } while ($sel);
        $this->redrawEditor();

    }

    /**
     * Ask for input
     *
     * @param string $prompt
     * @param string $value
     * @return string|null
     */
    public function ask(string $prompt, $value = ''): ?string
    {
        $plainPrompt = preg_replace('<\e\[.+?m>', '', $prompt);
        $promptLen = mb_strlen($plainPrompt);
        [$w,$h] = $this->term->getSize();

        $available = $w - $promptLen - 1;

        $prompting = true;
        $cursorPos = mb_strlen($value);
        $scrollPos = max(0, $cursorPos - $available);
        while ($prompting) {
            while (($cursorPos - $scrollPos) > $available) {
                $scrollPos++;
            }
            while (($cursorPos - $scrollPos) < 0) {
                $scrollPos--;
            }
            $cursorOffs = $cursorPos - $scrollPos;
            $this->term->setCursor(1, $h);
            echo "\e[0m".$prompt."\e[0m\e[K".mb_substr($value, $scrollPos, $available);
            $this->term->setCursor($promptLen+$cursorOffs+1, $h, true);
            while (null === ($ch = $this->term->readKey())) {
                usleep(10000);
            }
            if (mb_strlen($ch) == 1) {
                if (ord($ch) < 32) {
                    if (ord($ch) == 13) {
                        $this->term->setCursor(1, $h);
                        echo "\e[0m\e[K";
                        return $value;
                    }
                    if (ord($ch) == 3) {
                        // ctrl-c
                        $this->term->setCursor(1, $h);
                        echo "\e[0m\e[K";
                        return null;
                    }
                } elseif (ord($ch) == 127) {
                    if ($cursorPos > 0) {
                        $value = substr($value, 0, $cursorPos - 1) . substr($value, $cursorPos);
                        $cursorPos--;
                    }
                } else {
                    $value = mb_substr($value, 0, $cursorPos) . $ch . mb_substr($value, $cursorPos);
                    $cursorPos++;
                }
            } else {
                switch ($ch) {
                    case "k{LEFT}":
                        if ($cursorPos > 0)
                            $cursorPos--;
                        break;
                    case "k{RIGHT}":
                        if ($cursorPos < mb_strlen($value))
                            $cursorPos++;
                        break;
                    case "k{CLEFT}":
                        while ($cursorPos > 0) {
                            $cursorPos--;
                            if (mb_substr($value, $cursorPos - 1, 1) == " ") {
                                break;
                            }
                        }
                        break;
                    case "k{CRIGHT}":
                        $cursorPos = min(mb_strlen($value), (mb_strpos($value, " ", $cursorPos) ?: mb_strlen($value)) + 1);
                        break;
                    case "k{HOME}":
                        $cursorPos = 0;
                        break;
                    case "k{END}":
                        $cursorPos = mb_strlen($value);
                        break;
                }
            }
        }

    }

    /**
     * Redraw the editor
     *
     * @return void
     */
    public function redrawEditor()
    {
        [$w,$h] = $this->term->getSize();

        // Jump to the tail line if the cursor is past the end of the entry list
        if ($this->currentRow > count($this->list)) $this->currentRow = count($this->list);

        // Make sure the selection is in view
        while ($this->currentRow < $this->scrollOffset) $this->scrollOffset--;
        while ($this->currentRow > $h + $this->scrollOffset - 3) $this->scrollOffset++;

        // Nudge back so the tail line is visible but not selected
        if ($this->currentRow == count($this->list)) $this->currentRow--;

        $path = $this->list->getPathForIndex($this->currentRow);
        $node = $this->list->getNodeForIndex($this->currentRow);

        ob_start();

        $modified = $this->modified ? "\e[31m*\e[37m" : "";
        $this->term->setCursor(1, $h-1);
        echo "\e[40;37m\e[K{$modified}\e[1m{$this->shortfilename}\e[22m#\e[3m{$path}\e[37;23m";

        //$this->term->setCursor(1, $h);
        echo " = ";
        if ($node instanceof ValueNode) {
            echo json_encode($node->value, JSON_UNESCAPED_SLASHES|JSON_UNESCAPED_UNICODE);
        } elseif ($node instanceof ArrayNode) {
            echo "[…]";
        } elseif ($node instanceof ObjectNode) {
            echo "{…}";
        }
        echo "\e[0m";

        $this->redrawInfoBar();

        // $this->term->setCursor(1, $h);
        // echo "\e[0;37m\e[K";

        for ($n = 0; $n < $h - 2; $n++) {
            $this->list->drawEntry($n + 1, $n + $this->scrollOffset, $w, ($n + $this->scrollOffset == $this->currentRow));
        }
        // $this->term->setCursor(1, 1);
        // echo "\e[90;3m«Empty»\e[0m";

        ob_end_flush();
    }

    /**
     * Redraw the info bar
     *
     * @return void
     */
    private function redrawInfoBar(?array $keys = null)
    {
        [$w,$h] = $this->term->getSize();

        if (!$keys) {
            $keys = [
                'h' => 'Help',
                'D' => 'Delete',
            ];
            $node = $this->list->getNodeForIndex($this->currentRow);
            if ($node instanceof ValueNode) {
                $keys = [ ...$keys, 
                    'e' => 'Edit',
                ];
            }
            if ($node instanceof ObjectNode || $node instanceof ArrayNode) {
                $keys = [ ...$keys, 
                    'I' => 'Insert…',
                ];
            }
            if ($this->modified) {
                $keys = [ ...$keys, 
                    '^N' => 'New',
                    '^W' => 'Write',
                ];
            }
            $keys = [ ...$keys, 
                '^O' => 'Open',
                '^X' => 'Exit',
            ];
        }

        $this->term->setCursor(1, $h);
        echo "\e[37;40m\e[K";
        foreach ($keys as $key=>$info)
            echo "\e[2m\u{f104}\e[22;97;1m{$key}\e[22;37;2m\u{f105} \e[22;33m{$info}\e[37m ";
            //echo "\e[37;40;2m\u{e0b6}\e[7;37m{$key} \e[22m {$info} \e[27m\u{e0b4}\e[0m";
            //echo " \e[1m{$key}\e[2m \e[3m{$info}\e[0m \e[90m\u{2502}\e[0m";
    }

    private function showMessage(string $message): void
    {
        [$w,$h] = $this->term->getSize();
        $this->term->setCursor(1, $h);
        echo "\e[0m".$message."\e[K\e[0m";

    }

    private function setOpenedFilename(string $filename): void
    {
        $this->filename = $filename;
        $this->shortfilename = basename($filename);
        $this->setWindowTitle($this->shortfilename." - JSONEdit");
    }

    private function setWindowTitle(string $title): void
    {
        echo "\e]2;{$title}\x07";
    }

}

Editor::$helpText = <<<EOT
Welcome to JSONEdit! The editor you never knew you were missing!

# QuickStart

To get started, press I (shift-i) and add something to the document. Use the arrow keys to get around. To cancel a prompt, close a menu or close a dialog, press ctrl-C. When you are happy with your work, press ctrl-W and enter a filename to write. You can also press ctrl-O to browse for a file to open.

# Useful keys

### Main Window

   ↑↓  Navigate values in document
    h  Show this help
    H  Show information about the app and license
    i  Insert a new value
    I  Insert value, array or object
    e  Edit selected value
    E  Edit selected key
Enter  Edit string/number, toggle booleans
    D  Delete selected key
   ^D  Duplicate array item
    c  Toggle compact list view
    q  Toggle quoted keys in list view
    g  Toggle indentation guide lines visibility
    t  Toggle tail line visibility
   ^W  Write to a file
   ^R  Read from a file
   ^O  Open file browser
   ^N  New document with empty object
   ^X  Exit

### Dialogs/Menus

   ↑↓  Select/Scroll
Enter  Select value
   ^C  Cancel

# Doing stuff

### Adding keys or values
To add a key or a value, navigate to a value in an array or object, or to a specific array or object, and press "i". You will be prompted for the value, and for objects the key.
You can also press "I" to add arrays and objects. Just select what you want to add in the menu and press enter.

### Editing keys
You can edit keys on objects. For this, press "E".

### Editing values
To edit a value, press "e". The value is verbatim JSON, so strings should be quoted and all that. Anything that is unparsable JSON will be used as is, resulting in a string.

### Loading and Saving files
To load a file, press ^R and enter the filename to read. To write to a file, press ^W and enter the filename to write to.

### YAML or JSON?
There is no need to select YAML or JSON mode. All operations work the same, and the format is determined on load or save.

# Disclaimer
This is beta software, if not alpha. It kinda works, but there will be issues. Feel free to help out with a patch, or by filing bug reports.

# Support
Go to https://dev.noccylabs.info/noccy/jsonedit to find the source code, issue tracker, and learn more about the project!
EOT;
